/* initng-conf-gtk, the gtk configuration program for initng
 * Copyright (C) 2006 Daniel Malmgren <dm@mensa.se>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include "ngehandler.h"
#include "serviceparser.h"

#include <string.h>
#include <assert.h>
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <unistd.h>
#include <stdlib.h>

#include "initng-conf-gtk_globals.h"

void service_state(char *service, e_is is, int hidden, char *state, int update_screen) {

  /* lock all other threads */
  gdk_threads_enter();

  /* get, service_list that is a GtkTreeView in the main windows containing the list of all services,
   * defined in initng-conf-gtk.glade */
  GtkTreeView *service_list = GTK_TREE_VIEW(glade_xml_get_widget(xml, "service_list"));

  GtkStatusbar *statusbar = GTK_STATUSBAR(glade_xml_get_widget(xml, "statusbar"));
  guint context_id = gtk_statusbar_get_context_id(statusbar, "");
  
  /* The information about the stuff in the tree is stored in a model. On top of
   * that there's another model used for filtering, and on top of that there's
   * a model used for sorting. This means that to get the model we want (the one
   * that holds the data) we have to get the others first. Hence the following */
  GtkTreeModelSort *sortmodel = GTK_TREE_MODEL_SORT(gtk_tree_view_get_model(service_list));
  GtkTreeModelFilter *filtermodel = GTK_TREE_MODEL_FILTER(gtk_tree_model_sort_get_model(sortmodel));
  GtkTreeModel *service_model = gtk_tree_model_filter_get_model(filtermodel);

  /* Iterator used to point out the correct row added by add_node in service_model that
   * we are gonna modify. */
  GtkTreeIter iter;
  
  /* the real data storage struct from serviceparser.h */
  struct s_info *serviceinfo = NULL;

  /* if we had to add a new entry to the tree .. */
  /* add_node, defined in serviceparser.c, returns false if the node didn't exist */
  if(add_node(GTK_TREE_STORE(service_model), &iter, service)) {

    /* create som space for an s_info struct, using calloc to make sure allocated area is all NULL */
    serviceinfo = calloc(1,sizeof(struct s_info));

	/* Store the data in the tree. Since we had to create the node, we have no
	 * way of knowing what type it is, hence the "unknown". */
    gtk_tree_store_set(GTK_TREE_STORE(service_model), &iter, 1, _("unknown"), 
		       2, NULL, 4, serviceinfo, -1);
			   
	/* reset variable */
    serviceinfo->type = UNKNOWN;  /* todo, maby put UNKNOWN to 0 in the enum to mark it unset and remove this row? */
	
	/* insert the service name */
	serviceinfo->fullname = strdup(service);
  } else {
    /* Not newly allocated above, this is not known */
    gtk_tree_model_get(service_model, &iter, 4, &serviceinfo, -1);
  }
  
  /* serviceinfo shouldn't be null here, but if it is, it's really no need for panic */
  if(serviceinfo) {

    /* update those entrys get fron nge calling this function */
    serviceinfo->status_id = is;
    serviceinfo->hidden = hidden;
  
    /* set the state string to tree. When is state allocated?, shud it not be strdup(state)??
     * NOTE: state pointer is not valid after this function returns, freed by ngeclient_event_free(e); */
    gtk_tree_store_set(GTK_TREE_STORE(service_model), &iter, 2, strdup(state), -1);

    char *statustext = malloc(strlen(serviceinfo->fullname) + 13);
    strcpy(statustext, serviceinfo->fullname);

    /* put nice little colors in the tree depending on current state */
    switch(is) {
    case IS_UP:
      gtk_tree_store_set(GTK_TREE_STORE(service_model), &iter, 3, "#AAFFAA", -1);
      strcat(statustext, _(" is up."));
      break;
    case IS_DOWN:
      gtk_tree_store_set(GTK_TREE_STORE(service_model), &iter, 3, "#AAAAFF", -1);
      strcat(statustext, _(" is down"));
      break;
    case IS_FAILED:
      gtk_tree_store_set(GTK_TREE_STORE(service_model), &iter, 3, "#FFAAAA", -1);
      strcat(statustext, _(" has failed"));
      break;
    case IS_STARTING:
      gtk_tree_store_set(GTK_TREE_STORE(service_model), &iter, 3, "#FFFFAA", -1);
      strcat(statustext, _(" is starting"));
      break;
    case IS_STOPPING:
      gtk_tree_store_set(GTK_TREE_STORE(service_model), &iter, 3, "#AAFFFF", -1);
      strcat(statustext, _(" is stopping"));
      break;
    case IS_WAITING:
      gtk_tree_store_set(GTK_TREE_STORE(service_model), &iter, 3, "#FFAAFF", -1);
      strcat(statustext, _(" is waiting"));
      break;
    case IS_NEW:
      break;
    }
    gtk_statusbar_pop(statusbar, context_id);
    gtk_statusbar_push(statusbar, context_id, statustext);
    free(statustext);

    /* flush the db */
    gdk_flush();
  
    /* Only update screen on request */
    if(update_screen)
      gtk_tree_model_filter_refilter(filtermodel);
  }
  
  /* leave the thread lock */
  gdk_threads_leave();
}

/* when all initial_states for services are sent, this signal is set, and then
 * we can refilter and that will update screen */
void initial_state_finished(void) {
  gdk_threads_enter();
  GtkTreeView *service_list = GTK_TREE_VIEW(glade_xml_get_widget(xml, "service_list"));
  GtkTreeModelSort *sortmodel = GTK_TREE_MODEL_SORT(gtk_tree_view_get_model(service_list));
  GtkTreeModelFilter *filtermodel = GTK_TREE_MODEL_FILTER(gtk_tree_model_sort_get_model(sortmodel));

  GtkWidget *statusbar = glade_xml_get_widget(xml, "statusbar");
  guint context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), "");
  gtk_statusbar_pop(GTK_STATUSBAR(statusbar), context_id);

  /* Refilter will update the service tree */
  gtk_tree_model_filter_refilter(filtermodel);
  gdk_threads_leave();
}

void *nge_thread(void *args) {
  nge_connection *c = NULL;
  nge_event *e;

  /* open correct socket */
  c=ngeclient_connect(NGE_REAL);

  /* if open_socket fails, ngeclient_error is set */
  if (ngeclient_error) {
    fprintf(stderr, "ngeclient_error: ");
    fprintf(stderr, "%s\n", ngeclient_error);
    return NULL;
  }
  
  if(!c)
  {
  	fprintf(stderr, "NGE CONNECTION PROBLEMS\n");
	return NULL;
  }

  //Avoid nasty race scenarios
  sleep(1);

  /* continue for ever */
  while(c)
  {
  	
    /* get events, and wait for 999999999 microseconds */
  	e=get_next_event(c, 99999999);

 	/* if ngeclient_error is set - this is bad */
  	if (ngeclient_error) {
      fprintf(stderr, "ngeclient_error: ");
      fprintf(stderr, "%s\n", ngeclient_error);
    }
	
	/* and if it timed out, loop and try again */
	if(!e)
		continue;
	
	/* do for every event that comes in */
	switch (e->state_type)
	{
		case SERVICE_STATE_CHANGE:
			/* this will update screen on event */
			service_state(e->payload.service_state_change.service,
						   e->payload.service_state_change.is,
						   e->payload.service_state_change.hidden,
						   e->payload.service_state_change.state_name, 
						   1);
			break;
		case INITIAL_SERVICE_STATE_CHANGE:
			/* this will not, until initial_state_finished is sent */
			service_state(e->payload.service_state_change.service,
						   e->payload.service_state_change.is,
						   e->payload.service_state_change.hidden,
						   e->payload.service_state_change.state_name,
						   0);
			break;
		case INITIAL_STATE_FINISHED:
			initial_state_finished();
			break;
		
		/* on disconnect close the socket, and set c to null so the while loop stops */
		case DISSCONNECT:
 			ngeclient_close(c);
			c=NULL;
			break;
		default:
			break;
	}

	/* This will free all strings got */
	ngeclient_event_free(e);
  }

  /* check for failures */
  if (ngeclient_error) {
    fprintf(stderr, "%s\n", ngeclient_error);
  }

  /* clean up */
  ngeclient_close(c);

  return NULL;
}
